/**
 * @file libutouch-geis/geis.c
 * @brief implementation of the uTouch GEIS v2.0 API instance
 *
 * The GEIS API object encapsulates an entire persistent session with the
 * gesture recognition engine.  An application should be able to support
 * multiple simultaneous instances of the GEIS API abject (although it would be
 * foolish to design an application that way).
 */

/*
 * Copyright 2010, 2011 Canonical Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "geis_config.h"
#include "geis_private.h"

#include <errno.h>
#include "geis_atomic.h"
#include "geis_attr.h"
#include "geis_backend.h"
#include "geis_class.h"
#include "geis_device.h"
#include "geis_backend_multiplexor.h"
#include "geis_error.h"
#include "geis_event.h"
#include "geis_event_queue.h"
#include "geis_gesture_flick.h"
#include "geis_logging.h"
#include <stdarg.h>
#include <string.h>
#include <unistd.h>


/*
 * An internal struct to collect filterable attributes for facilities.
 */
typedef struct FilterableAttributeBag
{
  GeisFilterableAttribute store;
  GeisSize                size;
  GeisSize                count;
} *FilterableAttributeBag;

/*
 * An internal structure to track processing callbacks in some order.
 */
typedef struct GeisProcessingEntry *GeisProcessingEntry;

struct GeisProcessingEntry 
{
  GeisProcessingEntry     next;
  int                     priority;
  GeisProcessingCallback  event_callback;
  void                   *context;
};


/*
 * The API instance.
 */
struct _Geis
{
  GeisRefCount            refcount;
  GeisErrorStack          error_stack;
  GeisSubBag              subscription_bag;
  GeisBackendMultiplexor  backend_multiplexor;
  GeisBackend             backend;
  GeisEventQueue          input_event_queue;
  int                     input_event_signal_pipe[2];
  GeisProcessingEntry     processing_callbacks;
  GeisEventQueue          output_event_queue;
  GeisEventCallback       output_event_callback;
  void                   *output_event_callback_context;
  GeisEventCallback       class_event_callback;
  void                   *class_event_callback_context;
  FilterableAttributeBag  class_filterable_attributes;
  GeisGestureClassBag     gesture_classes;
  GeisEventCallback       device_event_callback;
  void                   *device_event_callback_context;
  FilterableAttributeBag  device_filterable_attributes;
  GeisDeviceBag           devices;
  FilterableAttributeBag  region_filterable_attributes;
  FilterableAttributeBag  special_filterable_attributes;
  GeisGestureFlick        flick;
};


/*
 * The default event callback -- just pushes events on the internal queue
 */
static void
_default_output_event_callback(Geis       geis,
                               GeisEvent  event,
                               void      *context GEIS_UNUSED)
{
  geis_debug("posting output event");
  geis_event_queue_enqueue(geis->output_event_queue, event);
}


/*
 * Handles device events coming in from the back end.
 */
static GeisBoolean
_device_event_handler(Geis geis, GeisEvent event)
{
  GeisBoolean handled = GEIS_FALSE;
  GeisAttr attr = NULL;

  attr = geis_event_attr_by_name(event, GEIS_EVENT_ATTRIBUTE_DEVICE);
  if (!attr)
  {
    geis_warning("invalid device event received from back end.");
    handled = GEIS_TRUE;
    goto final_exit;
  }
  
  GeisDevice device = geis_attr_value_to_pointer(attr);

  GeisEventType event_type = geis_event_type(event);
  if (event_type == GEIS_EVENT_DEVICE_AVAILABLE)
  {
    geis_device_bag_insert(geis->devices, device);
  }
  else if (event_type == GEIS_EVENT_DEVICE_UNAVAILABLE)
  {
    geis_device_bag_remove(geis->devices, device);
  }

  if (geis->device_event_callback != GEIS_DEFAULT_EVENT_CALLBACK)
  {
    geis->device_event_callback(geis, event, geis->device_event_callback_context);
    handled = GEIS_TRUE;
  }

final_exit:
  return handled;
}


/*
 * Handles class events coming in from the back end.
 */
static GeisBoolean
_class_event_handler(Geis geis, GeisEvent event)
{
  GeisBoolean handled = GEIS_FALSE;
  GeisAttr attr = NULL;

  attr = geis_event_attr_by_name(event, GEIS_EVENT_ATTRIBUTE_CLASS);
  if (!attr)
  {
    geis_warning("invalid class event received from back end.");
    handled = GEIS_TRUE;
    goto final_exit;
  }
  
  GeisGestureClass gesture_class = geis_attr_value_to_pointer(attr);

  GeisEventType event_type = geis_event_type(event);
  if (event_type == GEIS_EVENT_CLASS_AVAILABLE)
  {
    geis_gesture_class_bag_insert(geis->gesture_classes, gesture_class);
  }
  else if (event_type == GEIS_EVENT_CLASS_CHANGED)
  {
    /** @todo implement GEIS_EVENT_CLASS_CHANGED */
  }
  else if (event_type == GEIS_EVENT_CLASS_UNAVAILABLE)
  {
    geis_gesture_class_bag_remove(geis->gesture_classes, gesture_class);
  }

  if (geis->class_event_callback != GEIS_DEFAULT_EVENT_CALLBACK)
  {
    geis->class_event_callback(geis, event, geis->class_event_callback_context);
    handled = GEIS_TRUE;
  }

final_exit:
  return handled;
}


/*
 * Filters and transforms raw gesture events into cooked gesture events.
 */
static void
_input_event_handler(int fd, GeisBackendMultiplexorEvent mux_ev, void *context)
{
  Geis geis = (Geis)context;
  GeisProcessingEntry cb;

  if (mux_ev == GEIS_BE_MX_READ_AVAILABLE)
  {
    GeisEvent event;

    /* clear the input event signal */
    char buf[2];
    if (read(fd, buf, 1) != 1)
    {
      geis_warning("unexpected number of bytes read from signal pipe");
    }

    geis_debug("input event available");
    event = geis_event_queue_dequeue(geis->input_event_queue);
    if (event)
    {
      GeisBoolean handled = 0;
      switch (geis_event_type(event))
      {
	case GEIS_EVENT_DEVICE_AVAILABLE:
	case GEIS_EVENT_DEVICE_UNAVAILABLE:
	  handled = _device_event_handler(geis, event);
	  break;

	case GEIS_EVENT_CLASS_AVAILABLE:
	case GEIS_EVENT_CLASS_CHANGED:
	case GEIS_EVENT_CLASS_UNAVAILABLE:
	  handled = _class_event_handler(geis, event);
	  break;

	default:
	  break;
      }

      for (cb = geis->processing_callbacks; cb; cb = cb->next)
      {
	switch (cb->event_callback(event, cb->context))
	{
	  case GEIS_PROCESSING_DISPOSE_EVENT:
	    geis_event_delete(event);
	    goto final_exit;
	    break;
	  case GEIS_PROCESSING_COMPLETE:
	    goto processing_complete;
	    break;
	  case GEIS_PROCESSING_FAIL:
	    geis_warning("processig error in event handler");
	    break;
	  default:
	    break;
	}
      }

processing_complete:
      if (!handled)
      {
	geis->output_event_callback(geis,
	                            event,
	                            geis->output_event_callback_context);
      }
    }
  }

final_exit:
  return;
}


/*
 * Constructs a new filterable attribute bag.
 */
static FilterableAttributeBag
_filterable_attribute_bag_new()
{
  FilterableAttributeBag bag = calloc(1, sizeof(struct FilterableAttributeBag));
  if (!bag)
  {
    geis_error("failed to allocate filterable attribute bag");
    goto final_exit;
  }

  bag->size = 2;
  bag->count = 0;

  bag->store = calloc(1, sizeof(struct GeisFilterableAttribute));
  if (!bag)
  {
    geis_error("failed to allocate filterable attribute bag store");
    goto unwind_bag;
  }
  goto final_exit;

unwind_bag:
  free(bag);
  bag = NULL;
final_exit:
  return bag;
}


/*
 * Destroys a filterable attribute bag.
 */
static void
_filterable_attribute_bag_delete(FilterableAttributeBag bag)
{
  if (bag)
  {
    GeisSize i;
    for (i = 0; i < bag->count; ++i)
    {
      free((char *)bag->store[i].name);
    }
    free(bag->store);
  }
  free(bag);
}


static void
_filterable_attribute_copy(GeisFilterableAttribute src,
                           GeisFilterableAttribute dst)
{
  dst->name = strdup(src->name);
  dst->type = src->type;
  dst->add_term_callback = src->add_term_callback;
  dst->add_term_context  = src->add_term_context;
}


static void
_filterable_attribute_bag_insert(FilterableAttributeBag  bag,
                                 GeisFilterableAttribute fa)
{
  GeisSize new_count = bag->count + 1;
  if (new_count >= bag->size)
  {
    GeisSize new_size = bag->size * 2;
    GeisSize allocation_size = new_size * sizeof(struct GeisFilterableAttribute);
    GeisFilterableAttribute new_store = realloc(bag->store, allocation_size);
    if (!new_store)
    {
      geis_error("failed to reallocate filterable attribute bag store");
    }
    else
    {
      bag->store = new_store;
      bag->size = new_size;
    }
  }
  _filterable_attribute_copy(fa, &bag->store[bag->count]);
  bag->count = new_count;
}


/*
 * Applies the back end callback to the back end token for each filterable
 * attribute with a name matching the argument.
 */
static GeisStatus
_filterable_attribute_bag_foreach(FilterableAttributeBag  bag,
                                  GeisBackendToken        token,
                                  GeisString              name,
                                  GeisFilterOperation     op,
                                  void                   *value)
{
  GeisStatus status = GEIS_STATUS_SUCCESS;
  GeisSize i;
  for (i = 0; i < bag->count; ++i)
  {
    GeisFilterableAttribute fa = &bag->store[i];
    if (0 == strcmp(fa->name, name))
    {
      status = fa->add_term_callback(token, fa->add_term_context, name, op, value);
    }
  }
  return status;
}


/*
 * Runs filters for each regostered token in a facility.
 */
GeisStatus
geis_filterable_attribute_foreach(Geis geis,
                                  GeisFilterFacility  facility,
                                  GeisBackendToken    token,
                                  GeisString          name,
                                  GeisFilterOperation op,
                                  void               *value)
{
  GeisStatus status = GEIS_STATUS_UNKNOWN_ERROR;
  switch (facility)
  {
    case GEIS_FILTER_CLASS:
      status = _filterable_attribute_bag_foreach(geis->class_filterable_attributes,
                                                 token, name, op, value);
      break;
    case GEIS_FILTER_DEVICE:
      status = _filterable_attribute_bag_foreach(geis->device_filterable_attributes,
                                                 token, name, op, value);
      break;
    case GEIS_FILTER_REGION:
      status = _filterable_attribute_bag_foreach(geis->region_filterable_attributes,
                                                 token, name, op, value);
      break;
    case GEIS_FILTER_SPECIAL:
      status = _filterable_attribute_bag_foreach(geis->special_filterable_attributes,
                                                 token, name, op, value);
      break;
    default:
      break;
  }
  return status;
}


/**
 * Creates a new empty Geis API instance.
 */
static Geis
geis_new_empty()
{
  geis_error_clear(NULL);
  Geis geis = calloc(1, sizeof(struct _Geis));
  if (!geis)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("calloc failed");
    goto final_exit;
  }

  geis->subscription_bag = geis_subscription_bag_new(1);
  if (!geis->subscription_bag)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("creation of subscroption bag failed");
    free(geis);
    geis = NULL;
    goto unwind_geis;
  }

  geis->backend_multiplexor = geis_backend_multiplexor_new();
  if (!geis->backend_multiplexor)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("creation of back end multiplexor failed");
    goto unwind_subscription_bag;
  }

  geis->input_event_queue = geis_event_queue_new();
  if (!geis->input_event_queue)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("creation of input event queue failed");
    goto unwind_backend_mux;
  }
  if (pipe(geis->input_event_signal_pipe) < 0)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("error %d creating input event signal pipe: %s",
               errno, strerror(errno));
    goto unwind_input_queue;
  }
  geis_backend_multiplexor_add_fd(geis->backend_multiplexor,
                                  geis->input_event_signal_pipe[0],
                                  _input_event_handler,
                                  geis);

  geis->output_event_queue = geis_event_queue_new();
  if (!geis->output_event_queue)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("creation of output event queue failed");
    goto unwind_input_signal_pipe;
  }
  geis->output_event_callback = _default_output_event_callback;

  geis->class_filterable_attributes = _filterable_attribute_bag_new();
  if (!geis->class_filterable_attributes)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("creation of geis gesture class bag failed");
    goto unwind_output_queue;
  }

  geis->gesture_classes = geis_gesture_class_bag_new();
  if (!geis->gesture_classes)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("creation of geis gesture class bag failed");
    goto unwind_class_attrs;
  }
  geis->class_event_callback = _default_output_event_callback;

  geis->device_filterable_attributes = _filterable_attribute_bag_new();
  if (!geis->device_filterable_attributes)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("creation of geis device bag failed");
    goto unwind_class_bag;
  }

  geis->devices = geis_device_bag_new();
  if (!geis->devices)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("creation of geis device bag failed");
    goto unwind_device_attrs;
  }
  geis->device_event_callback = _default_output_event_callback;

  geis->region_filterable_attributes = _filterable_attribute_bag_new();
  if (!geis->region_filterable_attributes)
  {
    goto unwind_device_bag;
  }

  geis->special_filterable_attributes = _filterable_attribute_bag_new();
  if (!geis->special_filterable_attributes)
  {
    goto unwind_region_attrs;
  }

  goto final_exit;

unwind_region_attrs:
  _filterable_attribute_bag_delete(geis->region_filterable_attributes);
unwind_device_bag:
  geis_device_bag_delete(geis->devices);
unwind_device_attrs:
  _filterable_attribute_bag_delete(geis->device_filterable_attributes);
unwind_class_bag:
  geis_gesture_class_bag_delete(geis->gesture_classes);
unwind_class_attrs:
  _filterable_attribute_bag_delete(geis->class_filterable_attributes);
unwind_output_queue:
  geis_event_queue_delete(geis->output_event_queue);
unwind_input_signal_pipe:
  close(geis->input_event_signal_pipe[0]);
  close(geis->input_event_signal_pipe[1]);
unwind_input_queue:
  geis_event_queue_delete(geis->input_event_queue);
unwind_backend_mux:
  geis_backend_multiplexor_delete(geis->backend_multiplexor);
unwind_subscription_bag:
  geis_subscription_bag_delete(geis->subscription_bag);
unwind_geis:
  free(geis);
  geis = NULL;

final_exit:
  return geis;
}


typedef enum _BackendType
{
  BACK_END_TYPE_NONE,
  BACK_END_TYPE_MOCK_ENGINE,
  BACK_END_TYPE_XCB
} BackendType;

/**
 * Sets optional parts of a Geis API instance from a variable argument list.
 */
static GeisBoolean
_set_valist(Geis geis, GeisString init_arg_name, va_list varargs)
{
  GeisBoolean status = GEIS_TRUE;
  BackendType back_end_type = BACK_END_TYPE_NONE;
  GeisBoolean track_devices = GEIS_FALSE;
  GeisBoolean track_classes = GEIS_FALSE;

  while (init_arg_name)
  {
    if (0 == strcmp(init_arg_name, GEIS_INIT_SERVICE_PROVIDER))
    {
      geis_debug("initializing GEIS server");
    }
    else if (0 == strcmp(init_arg_name, GEIS_INIT_TRACK_DEVICES))
    {
      track_devices = GEIS_TRUE;
    }
    else if (0 == strcmp(init_arg_name, GEIS_INIT_TRACK_GESTURE_CLASSES))
    {
      track_classes = GEIS_TRUE;
    }
    else if (0 == strcmp(init_arg_name, GEIS_INIT_UTOUCH_MOCK_ENGINE))
    {
      if (back_end_type != BACK_END_TYPE_NONE)
      {
	geis_error("multiple back ends requested, only using last request");
      }
      back_end_type = BACK_END_TYPE_MOCK_ENGINE;
    }
    else if (0 == strcmp(init_arg_name, GEIS_INIT_UTOUCH_XCB))
    {
      if (back_end_type != BACK_END_TYPE_NONE)
      {
	geis_error("multiple back ends requested, only using last request");
      }
      back_end_type = BACK_END_TYPE_XCB;
    }

    init_arg_name = va_arg(varargs, GeisString);
  }

  if (back_end_type == BACK_END_TYPE_MOCK_ENGINE)
  {
    geis->backend = geis_backend_by_name(geis, GEIS_INIT_UTOUCH_MOCK_ENGINE);
  }
  else if (back_end_type ==BACK_END_TYPE_XCB)
  {
    geis->backend = geis_backend_by_name(geis, GEIS_INIT_UTOUCH_XCB);
  }
  else
  {
    geis_warning("back end not specified, defaulting to XCB");
    geis->backend = geis_backend_by_name(geis, GEIS_INIT_UTOUCH_XCB);
  }
  if (!geis->backend)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("can not create back end");
    status = GEIS_FALSE;
  }

  return status;
}


/**
 * Creates an initialized Geis API instance.
 */
Geis
geis_new(GeisString init_arg_name, ...)
{
  GeisBoolean success = GEIS_FALSE;
  Geis geis = geis_new_empty();
  if (geis)
  {
    geis_ref(geis);

    va_list varargs;
    va_start(varargs, init_arg_name);
    success = _set_valist(geis, init_arg_name, varargs);
    va_end(varargs);
  }

  if (!success)
  {
    geis_error_push(NULL, GEIS_STATUS_UNKNOWN_ERROR);
    geis_error("can not initialize GEIS API");
    geis_delete(geis);
    geis = NULL;
    goto final_exit;
  }
  geis->flick = geis_gesture_flick_new(geis);

final_exit:
  return geis;
}


static void _geis_destroy(Geis geis)
{
  geis_debug("begins");
  geis_gesture_flick_delete(geis->flick);
  _filterable_attribute_bag_delete(geis->special_filterable_attributes);
  _filterable_attribute_bag_delete(geis->region_filterable_attributes);
  geis_device_bag_delete(geis->devices);
  _filterable_attribute_bag_delete(geis->device_filterable_attributes);
  geis_gesture_class_bag_delete(geis->gesture_classes);
  _filterable_attribute_bag_delete(geis->class_filterable_attributes);
  geis_event_queue_delete(geis->output_event_queue);
  close(geis->input_event_signal_pipe[0]);
  close(geis->input_event_signal_pipe[1]);
  geis_event_queue_delete(geis->input_event_queue);
  if (geis->backend)
    geis_backend_delete(geis->backend);
  geis_backend_multiplexor_delete(geis->backend_multiplexor);
  geis_subscription_bag_delete(geis->subscription_bag);
  free(geis);
}

/*
 * Disposes of a Geis API instance.
 */
GeisStatus
geis_delete(Geis geis)
{
  if (geis == NULL)
  {
    return GEIS_STATUS_BAD_ARGUMENT;
  }

  geis_subscription_bag_invalidate(geis->subscription_bag);
  geis_unref(geis);
  return GEIS_STATUS_SUCCESS;
}

/*
 * Increases the reference count of an API instance object.
 */
Geis
geis_ref(Geis geis)
{
  geis_atomic_ref(&geis->refcount);
  return geis;
}


/*
 * Decremenets the reference count of an API instance object.
 */
void
geis_unref(Geis geis)
{
  if (0 == geis_atomic_unref(&geis->refcount))
  {
    _geis_destroy(geis);
  }
}


/**
 * Gets a named configuration item.
 */
GeisStatus
geis_get_configuration(Geis        geis, 
                       GeisString  configuration_item_name,
                       void       *configuration_item_value)
{
  GeisStatus status = GEIS_STATUS_NOT_SUPPORTED;

  if (0 == strcmp(configuration_item_name, GEIS_CONFIGURATION_FD))
  {
    *(int*)configuration_item_value
        = geis_backend_multiplexor_fd(geis->backend_multiplexor);
    status = GEIS_STATUS_SUCCESS; 
  }
  else if (0 == strcmp(configuration_item_name, GEIS_CONFIG_UTOUCH_MAX_EVENTS))
  {
    *(int*)configuration_item_value
        = geis_backend_multiplexor_max_events_per_pump(geis->backend_multiplexor);
    status = GEIS_STATUS_SUCCESS; 
  }

  return status;
}


/*
 * Sets a named configuration item.
 */
GeisStatus
geis_set_configuration(Geis        geis,
                       GeisString  configuration_item_name,
                       void       *configuration_item_value)
{
  GeisStatus status = GEIS_STATUS_NOT_SUPPORTED;

  if (0 == strcmp(configuration_item_name, GEIS_CONFIG_UTOUCH_MAX_EVENTS))
  {
    int max_events = *(int *)configuration_item_value;
    geis_backend_multiplexor_set_max_events_per_pump(geis->backend_multiplexor,
                                                     max_events);
    status = GEIS_STATUS_SUCCESS; 
  }
  return status;
}


/*
 * Registers a callback to receive device change notifications.
 */
void
geis_register_device_callback(Geis               geis,
                              GeisEventCallback  event_callback,
                              void              *context)
{
  geis->device_event_callback = event_callback;
  geis->device_event_callback_context = context;
}


/*
 * Registers a callback to receive gesture class change notifications.
 */
void
geis_register_class_callback(Geis               geis,
                             GeisEventCallback  event_callback,
                             void              *context)
{
  geis->class_event_callback = event_callback;
  geis->class_event_callback_context = context;
}


/*
 * Registers an event-processing callback.
 */
void
geis_register_processing_callback(Geis                    geis,
                                  int                     priority,
                                  GeisProcessingCallback  event_callback,
                                  void                   *context)
{
  GeisProcessingEntry new_entry = calloc(1, sizeof(struct GeisProcessingEntry));
  new_entry->priority = priority;
  new_entry->event_callback = event_callback;
  new_entry->context = context;

  if (!geis->processing_callbacks)
  {
    geis->processing_callbacks = new_entry;
  }
  else
  {
    GeisProcessingEntry cb, prev;
    for (cb = geis->processing_callbacks; cb; cb = cb->next)
    {
      if (priority < cb->priority)
      {
	if (cb == geis->processing_callbacks)
	{
	  geis->processing_callbacks = new_entry;
	}
	else
	{
	  new_entry->next = prev->next;
	  prev->next = new_entry;
	}
	break;
      }
      prev = cb;
    }
    if (!cb)
    {
      prev->next = new_entry;
    }
  }
}


/*
 * Registers an application-supplied event callback.
 */
void 
geis_register_event_callback(Geis               geis,
                             GeisEventCallback  output_event_callback,
                             void              *context)
{
  if (output_event_callback == GEIS_DEFAULT_EVENT_CALLBACK)
  {
    geis->output_event_callback = _default_output_event_callback;
  }
  else
  {
    geis->output_event_callback = output_event_callback;
  }
  geis->output_event_callback_context = context;
}


/*
 * Pumps the GEIS v2 event loop.
 */
GeisStatus
geis_dispatch_events(Geis geis)
{
  GeisStatus status = geis_backend_multiplexor_pump(geis->backend_multiplexor);
  return status;
}


/*
 * Posts an event through the API.
 *
 * Pushes the new event onto the input event queue and signals that a new event
 * has arrived.
 */
void
geis_post_event(Geis geis, GeisEvent event)
{
  geis_event_queue_enqueue(geis->input_event_queue, event);
  if (write(geis->input_event_signal_pipe[1], "1", 1) != 1)
  {
    geis_error("error %d writing input event signal: %s", errno, strerror(errno));
  }
}


/*
 * Pulls the next event off the queue.
 */
GeisStatus
geis_next_event(Geis geis, GeisEvent *event)
{
  GeisStatus status = GEIS_STATUS_EMPTY;
  *event = geis_event_queue_dequeue(geis->output_event_queue);
  if (*event)
  {
    status = geis_event_queue_is_empty(geis->output_event_queue)
           ? GEIS_STATUS_SUCCESS
           : GEIS_STATUS_CONTINUE;
  }

  return status;
}


/*
 * Adds a back end file descriptor to multiplex.
 */
void
geis_multiplex_fd(Geis                        geis,
                  int                         fd,
                  GeisBackendFdEventCallback  callback,
                  void                       *context)
{
  geis_backend_multiplexor_add_fd(geis->backend_multiplexor,
                                  fd, callback, context);
}


/*
 * Removes a back end file descriptor from the multiplex.
 */
void
geis_demultiplex_fd(Geis geis, int fd)
{
  geis_backend_multiplexor_remove_fd(geis->backend_multiplexor, fd);
}


GeisErrorStack *
geis_error_stack(Geis geis)
{
  return &geis->error_stack;
}


GeisSubBag
geis_subscription_bag(Geis geis)
{
  return geis->subscription_bag;
}


/*
 * Creates a new backend token.
 */
GeisBackendToken
geis_backend_token_new(Geis geis, GeisBackendTokenInitState init_state)
{
  return geis_backend_create_token(geis->backend, init_state);
}


/*
 * Resgisters a new gesture class with the API.
 */
void
geis_register_gesture_class(Geis                    geis,
                            GeisGestureClass        gesture_class,
                            GeisSize                filterable_attr_count,
                            GeisFilterableAttribute filterable_attrs)
{
  GeisAttr attr = NULL;
  GeisEvent event = NULL;
  GeisSize i;

  /* Add the filterable attributes. */
  for (i = 0; i < filterable_attr_count; ++i)
  {
    _filterable_attribute_bag_insert(geis->class_filterable_attributes,
                                     &filterable_attrs[i]);
  }

  geis_debug("registering class \"%s\" id %d",
             geis_gesture_class_name(gesture_class),
             geis_gesture_class_id(gesture_class));

  /* Add a "new gesture class" event to the processing queue. */
  event = geis_event_new(GEIS_EVENT_CLASS_AVAILABLE);
  attr = geis_attr_new(GEIS_EVENT_ATTRIBUTE_CLASS,
                       GEIS_ATTR_TYPE_POINTER,
                       gesture_class);
  geis_event_add_attr(event, attr);
  geis_post_event(geis, event);
}


GeisGestureClassBag
geis_gesture_classes(Geis geis)
{
  return geis->gesture_classes;
}


/*
 * Resgisters a new input devicve with the API.
 */
void
geis_register_device(Geis                    geis,
                     GeisDevice              device,
                     GeisSize                filterable_attr_count,
                     GeisFilterableAttribute filterable_attrs)
{
  GeisAttr attr = NULL;
  GeisEvent event = NULL;
  GeisSize i;

  geis_debug("registering device \"%s\" id %d",
             geis_device_name(device),
             geis_device_id(device));

  /* Add the filterable attributes. */
  for (i = 0; i < filterable_attr_count; ++i)
  {
    _filterable_attribute_bag_insert(geis->device_filterable_attributes,
                                     &filterable_attrs[i]);
  }

  /* Add a "new device" event to the processing queue. */
  event = geis_event_new(GEIS_EVENT_DEVICE_AVAILABLE);
  attr = geis_attr_new(GEIS_EVENT_ATTRIBUTE_DEVICE,
                       GEIS_ATTR_TYPE_POINTER,
                       device);
  geis_event_add_attr(event, attr);
  geis_post_event(geis, event);
}


GeisDeviceBag
geis_devices(Geis geis)
{
  return geis->devices;
}


/*
 * Resgisters a new region with the API.
 */
void
geis_register_region(Geis                    geis,
                     GeisRegion              region GEIS_UNUSED,
                     GeisSize                filterable_attr_count,
                     GeisFilterableAttribute filterable_attrs)
{
  GeisSize i;

  geis_debug("registering region");

  /* Add the filterable attributes. */
  for (i = 0; i < filterable_attr_count; ++i)
  {
    _filterable_attribute_bag_insert(geis->region_filterable_attributes,
                                     &filterable_attrs[i]);
  }
}


/*
 * Resgisters a new special filter with the API.
 */
void
geis_register_special(Geis                    geis,
                      GeisSize                filterable_attr_count,
                      GeisFilterableAttribute filterable_attrs)
{
  GeisSize i;

  geis_debug("registering feature");

  /* Add the filterable attributes. */
  for (i = 0; i < filterable_attr_count; ++i)
  {
    _filterable_attribute_bag_insert(geis->special_filterable_attributes,
                                     &filterable_attrs[i]);
  }
}


GeisAttrType
geis_get_device_attr_type(Geis geis, GeisString attr_name)
{
  GeisSize i;
  for (i = 0; i < geis->device_filterable_attributes->count; ++i)
  {
    GeisFilterableAttribute fa = &geis->device_filterable_attributes->store[i];
    if (0 == strcmp(attr_name, fa->name))
    {
      return fa->type;
    }
  }
  return GEIS_ATTR_TYPE_UNKNOWN;
}


GeisAttrType
geis_get_class_attr_type(Geis geis, GeisString attr_name)
{
  GeisSize i;
  for (i = 0; i < geis->class_filterable_attributes->count; ++i)
  {
    GeisFilterableAttribute fa = &geis->class_filterable_attributes->store[i];
    if (0 == strcmp(attr_name, fa->name))
    {
      return fa->type;
    }
  }
  return GEIS_ATTR_TYPE_UNKNOWN;
}


GeisAttrType
geis_get_region_attr_type(Geis geis, GeisString attr_name)
{
  GeisSize i;
  for (i = 0; i < geis->region_filterable_attributes->count; ++i)
  {
    GeisFilterableAttribute fa = &geis->region_filterable_attributes->store[i];
    if (0 == strcmp(attr_name, fa->name))
    {
      return fa->type;
    }
  }
  return GEIS_ATTR_TYPE_UNKNOWN;
}



GeisAttrType
geis_get_special_attr_type(Geis geis, GeisString attr_name)
{
  GeisSize i;
  for (i = 0; i < geis->special_filterable_attributes->count; ++i)
  {
    GeisFilterableAttribute fa = &geis->special_filterable_attributes->store[i];
    if (0 == strcmp(attr_name, fa->name))
    {
      return fa->type;
    }
  }
  return GEIS_ATTR_TYPE_UNKNOWN;
}


